import astropy.units as u
import ogameasure

from ...core.security import busy
from ...utils import skip_on_simulator
from .attenuator_base import NetworkAttenuator


class A11713B(NetworkAttenuator):
    """Attenuator, which can attennuate IF sigal power.

    Notes
    -----

    Configuration items for this device:

    host : str
        IP address for GPIB communicator.

    port : int
        GPIB port of using devices.

    channel : Dict[str]
        Human-readable channel name. The value should be
        mapping from human readableversion (str) to
        device level identifier (int). You can assign any name to the
        channels up to two channels: "X", "Y".
        For example: `{ 2R = X, 2L = Y}`

    """

    Manufacturer = "Agilent"
    Model = "11713B"

    Identifier = "host"

    @skip_on_simulator
    def __init__(self) -> None:
        com = ogameasure.gpib_prologix(host=self.Config.host, gpibport=self.Config.port)
        self.io = ogameasure.Agilent.agilent_11713B(com)

    def get_loss(self, id: str) -> u.Quantity:
        with busy(self, "busy"):
            ch = self.Config.channel[id]
            try:
                return self.io.att_level_query(ch) * u.dB
            except IndexError:
                pass
            raise ValueError(f"Invalid channel: {ch}")

    def set_loss(self, dB: int, id: str) -> None:
        with busy(self, "busy"):
            ch = self.Config.channel[id]
            self.io.att_level_set(dB, ch)

    def finalize(self) -> None:
        self.io.com.close()



from ...core.security import busy
from .attenuator_base import CurrentAttenuator


class CPZ340516(CurrentAttenuator):
    """LOattenuator, which can convert by 8 channels.

    Notes
    -----
    Configuration items for this device:

    channel : {{100GHz = 1}}

    range : {"DA0_100mA"} or {"DA0_1mA"}
        If DA0_100mA, 1-100mA is possible. Else if DA0_1mA, 1-10mA is possible.
        Default value is DA0_100mA.

    rate : {0.1}

    rsw_id : {0, 1, ..., 16} or {"0", "1", ..., "9", "A", ..., "F"}
        Board identifier. This should be set to the same value as the rotary switch
        "RSW1" mounted on the side of the board. The board is shipped with default RSW1
        setting of 0. This ID would be non-zero, when multiple PCI board of same model
        are mounted on a single FA (Factory Automation) controller.

    See defaults setting file in ``neclib/defaults/config.toml``.

    """

    Manufacturer = "Interface"
    Model = "CPZ340516"

    Identifier = "rsw_id"

    def __init__(self):
        import pyinterface

        self.rsw_id = self.Config.rsw_id
        self.range = self.Config.range
        self.channel = self.Config.channel
        self.io = pyinterface.open(3405, self.rsw_id)
        for i in self.channel.values():
            try:
                self.io.set_outputrange(i, self.range)
            except Exception as e:
                if e == "lspci: Unable to load libkmod resources: error -2":
                    pass
                else:
                    raise ValueError(f"Invalid channel: {i}")

    def get_outputrange(self, id: int) -> dict:
        ch = self.Config.channel[id]
        with busy(self, "busy"):
            return self.io.get_outputrange(ch)

    def set_outputrange(self, id: int, outputrange: str):
        self.io.set_outputrange(id, outputrange)

    def output_current(self, id: int, current: float):
        ch = self.Config.channel[id]
        with busy(self, "busy"):
            self.io.output_current(ch, current)

    def finalize(self) -> None:
        self.io.finalize()



__all__ = ["CPZ2724"]

import time

from ... import get_logger, utils
from .motor_base import Motor


class CPZ2724(Motor):
    # """Digital I/O board for membrane opener of NANTEN2.
    # Notes
    # -----
    # Configuration items for this device:
    # rsw_id : {0, 1, ..., 16} or {"0", "1", ..., "9", "A", ..., "F"}
    #     Board identifier. This should be set to the same value as the rotary switch
    #     "RSW1" mounted on the side of the board. The board is shipped with default
    #     RSW1 setting of 0. This ID would be non-zero, when multiple PCI board of same
    #     model are mounted on a single FA (Factory Automation) controller.
    # See defaults setting file in neclib/defaults/config.toml.
    # """

    Manufacturer = "Interface"
    Model = "CPZ2724"

    Identifier = "rsw_id"

    def __init__(self) -> None:
        self.logger = get_logger(self.__class__.__name__)
        self.rsw_id = self.Config.rsw_id

        self.speed_to_rate = float((7 / 12) * (10000 / 3600))

        self.io = self._initialize_io()

    @utils.skip_on_simulator
    def _initialize_io(self):
        import pyinterface

        io = pyinterface.open(2724, self.rsw_id)
        if io is None:
            raise RuntimeError("Cannot communicate with the CPZ board.")

        io.initialize()

        return io

    # Antenna Control

    def set_speed(
        self,
        speed: float,
        axis: str,
    ):
        speed_float = float(speed) * self.speed_to_rate
        self.antenna_move(int(speed_float), axis)
        return

    def get_speed(self, axis: str) -> float:
        word = None
        if axis == "az":
            word = "IN1_16"
        elif axis == "el":
            word = "IN17_32"
        else:
            raise ValueError(f"No valid axis : {axis}")
        speed = self.io.input_word(word)
        return speed

    def antenna_move(self, speed: int, axis: str) -> None:
        n_bits = 16
        word = None
        if axis == "az":
            word = "OUT1_16"
        elif axis == "el":
            word = "OUT17_32"
        else:
            raise ValueError(f"No valid axis : {axis}")

        cmd = bin(speed)[2:].zfill(n_bits)[::-1]  # [::-1] for little endian.
        cmd = [int(char) for char in cmd]
        self.io.output_word(word, cmd)

    def antenna_stop(self):
        self.antenna_move(0, "az")
        self.antenna_move(0, "el")

    def antenna_status(self) -> dict[str, str]:
        status_dict = {}
        for i in ["az", "el"]:
            status = self.get_speed(i)
            if status.bytes == "0x0000":
                antenna_status = "STOP"
            else:
                antenna_status = "MOVE"
            status_dict[i] = antenna_status
        return status_dict

    def set_step(self, step: int, axis: str) -> None:
        pass

    def get_step(self, axis: str) -> None:
        pass

    # Dome Control

    def dome_move(self, speed: str, turn: str):
        buffer = [0, 1, 0, 0]
        if turn == "right":
            buffer[0] = 0
        elif turn == "left":
            buffer[0] = 1
        else:
            raise ValueError(f"No valid turn direction : {turn}")
        if speed == "low":
            buffer[2:4] = [0, 0]
        elif speed == "mid":
            buffer[2:4] = [1, 0]
        elif speed == "high":
            buffer[2:4] = [0, 1]
        else:
            raise ValueError(f"No valid speed : {speed}")
        self.io.output_point(buffer, 1)

    def dome_oc(self, pos: str) -> None:
        # posにはopen or close を入れる
        ret = self.dome_status()
        if ret[1] != pos and ret[3] != pos:
            buff = self.Config.position[pos.lower()]
            self.io.output_point(buff, 5)
            while ret[1] != pos and ret[3] != pos:
                time.sleep(5)
                ret = self.dome_status()
        buff = [0, 0]
        self.io.output_point(buff, 5)

    def dome_stop(self) -> None:
        buff = [0]
        self.io.output_point(buff, 2)

    def dome_fan(self, fan: str) -> None:
        # fanにはon or off を入れる
        if fan == "on":
            fan_bit = [1, 1]
            self.io.output_point(fan_bit, 9)
        else:
            fan_bit = [0, 0]
            self.io.output_point(fan_bit, 9)

    def dome_status(self):
        ret = self.io.input_point(2, 6)
        if ret[0] == 0:
            self.right_act = "OFF"
        else:
            self.right_act = "DRIVE"

        if ret[1] == 0:
            if ret[2] == 0:
                self.right_pos = "MOVE"
            else:
                self.right_pos = "CLOSE"
        else:
            self.right_pos = "OPEN"

        if ret[3] == 0:
            self.left_act = "OFF"
        else:
            self.left_act = "DRIVE"

        if ret[4] == 0:
            if ret[5] == 0:
                self.left_pos = "MOVE"
            else:
                self.left_pos = "CLOSE"
        else:
            self.left_pos = "OPEN"
        return [self.right_act, self.right_pos, self.left_act, self.left_pos]

    # Membrane Control

    def memb_oc(self, pos: str) -> None:
        # posには OPEN or CLOSE を入れる
        ret = self.memb_status()
        if ret[1] != pos:
            buff = self.Config.position[pos.lower()]
            self.io.output_point(buff, 7)
            while ret[1] != pos:
                time.sleep(5)
                ret = self.memb_status()
        buff = [0, 0]
        self.io.output_point(buff, 7)

    def memb_status(self) -> list[str, str]:
        ret = self.io.input_point(8, 3)
        if ret[0] == 0:
            self.memb_act = "OFF"
        else:
            self.memb_act = "DRIVE"

        if ret[1] == 0:
            if ret[2] == 0:
                self.memb_pos = "MOVE"
            else:
                self.memb_pos = "CLOSE"
        else:
            self.memb_pos = "OPEN"
        return [self.memb_act, self.memb_pos]

    # M2 Control

    def m2_move(self, dist: int):
        status = self.m2_status()
        puls = self.um_to_puls(dist, status)
        if puls is None:
            raise ValueError("Please change distance.")
        self.MoveIndexFF(puls)

    def m2_status(self) -> list:
        buff = []
        buff2 = []
        in1_8 = self.io.input_byte("IN1_8").to_list()
        in9_16 = self.io.input_byte("IN9_16").to_list()

        if in9_16[6] == 1:  # [0,0,0,0,0,0,1,0]:
            m_limit_up = 1
        else:
            m_limit_up = 0
        if in9_16[7] == 1:  # [0,0,0,0,0,0,0,1]:
            m_limit_down = 1
        else:
            m_limit_down = 0

        for i in range(8):
            if i != 0 and in1_8 == [0, 0, 0, 0, 0, 0, 0, 0]:
                buff.append(0)
            if in1_8[i] == 0:
                buff.append(0)
            else:
                buff.append(1)

        for i in range(8):
            if i != 0 and in9_16 == [0, 0, 0, 0, 0, 0, 0, 0]:
                buff2.append(0)
            if in9_16[i] == 0:
                buff2.append(0)
            else:
                buff2.append(1)

        # calculate each digit
        total = (
            buff[0] * 1
            + buff[1] * 2
            + buff[2] * pow(2.0, 2.0)
            + buff[3] * pow(2.0, 3.0)
        ) / 100.0
        total = (
            total
            + (
                buff[4] * 1
                + buff[5] * 2
                + buff[6] * pow(2.0, 2.0)
                + buff[7] * pow(2.0, 3.0)
            )
            / 10.0
        )
        total2 = (
            buff2[0] * 1
            + buff2[1] * 2
            + buff2[2] * pow(2.0, 2.0)
            + buff2[3] * pow(2.0, 3.0)
        )
        total2 = total2 + (buff2[4] * 1) * 10

        m_pos = (total + total2) * pow(-1.0, (buff2[5] + 1))
        return [m_pos, m_limit_up, m_limit_down]

    def um_to_puls(self, dist: int, status: list[int]) -> int:
        self.pulse_rate = self.Config.pulse_rate

        puls = int(dist) * self.pulse_rate
        if (
            dist / 1000.0 + float(status[0]) <= -4.0
            or dist / 1000.0 + float(status[0]) >= 5.5
        ):
            print("move limit")
            return
        if status[1] == 0 and puls < 0:
            print("can't move up direction")
            return
        if status[2] == 0 and puls > 0:
            print("can't move down direction")
            return
        return puls

    def MoveIndexFF(self, puls: int):
        self.cw = self.Config.cw
        self.ccw = self.Config.ccw
        self.motor_speed = self.Config.motor_speed

        if puls >= -65535 and puls <= 65535:
            # index mode
            self.io.output_byte("OUT1_8", [0, 0, 0, 1, 0, 0, 0, 0])
            self.Strobe()
            # step no.
            self.io.output_byte("OUT1_8", [1, 1, 1, 1, 1, 1, 1, 1])
            self.Strobe()
            # position set
            self.io.output_byte("OUT1_8", [0, 0, 0, 0, 0, 0, 1, 1])
            self.Strobe()
            # direction
            if puls >= 0:
                self.io.output_byte("OUT1_8", self.cw, fmt="<I")
                self.Strobe()
            else:
                self.io.output_byte("OUT1_8", self.ccw, fmt="<I")
                self.Strobe()
            # displacement
            self.io.output_byte("OUT1_8", [0, 0, 0, 0, 0, 0, 0, 0])
            self.Strobe()
            self.io.output_byte("OUT1_8", int(abs(puls) / 256), fmt="<I")
            self.Strobe()
            self.io.output_byte("OUT1_8", int(abs(puls) % 256), fmt="<I")
            self.Strobe()
            # start
            self.io.output_byte("OUT1_8", [0, 0, 0, 1, 1, 0, 0, 0])
            self.Strobe()
            time.sleep((abs(puls) / self.motor_speed / 10.0) + 1.0)
            print("Motor stopped")
        else:
            print("Puls number is over.")
            print("Please command x : 10*x [um]")
            return False
        return True

    def Strobe(self):
        time.sleep(0.01)
        self.io.output_byte("OUT9_16", [1, 0, 0, 0, 0, 0, 0, 0])
        time.sleep(0.01)
        self.io.output_byte("OUT9_16", [0, 0, 0, 0, 0, 0, 0, 0])
        time.sleep(0.01)
        return

    def finalize(self) -> None:
        # dome stop
        self.io.output_point([0], 2)



from abc import abstractmethod

import astropy.units as u

from ..device_base import DeviceBase


class ADConverter(DeviceBase):

    Manufacturer: str = ""
    Model: str

    @abstractmethod
    def get_voltage(self, id: str) -> u.Quantity: ...

    @abstractmethod
    def get_current(self, id: str) -> u.Quantity: ...

    @abstractmethod
    def finalize(self) -> None: ...



__all__ = ["CPZ6204"]

from ... import utils, get_logger
from .encoder_base import Encoder
from astropy import units as u


class CPZ6204(Encoder):
    # """Encoder readout.

    # Notes
    # -----
    # Configuration items for this device:
    # rsw_id : {0, 1, ..., 16} or {"0", "1", ..., "9", "A", ..., "F"}
    #     Board identifier. This should be set to the same value as the rotary switch
    #     "RSW1" mounted on the side of the board. The board is shipped with default
    #     RSW1 setting of 0. This ID would be non-zero, when multiple PCI board of same
    #     model are mounted on a single FA (Factory Automation) controller.
    # See defaults setting file in neclib/defaults/config.toml.
    # """

    Manufacturer = "Interface"
    Model = "CPZ6204"

    Identifier = "rsw_id"

    def __init__(self) -> None:
        self.logger = get_logger(self.__class__.__name__)
        self.rsw_id = self.Config.rsw_id

        self.enc_Az = 0
        self.enc_El = 45 * 3600
        self.resolution = 360 * 3600 / (23600 * 400)

        self.io = self._initialize()

    @utils.skip_on_simulator
    def _initialize(self):
        import pyinterface

        io = pyinterface.open(6204, self.rsw_id)
        if io is None:
            raise RuntimeError("Cannot communicate with the CPZ board.")
        mode = io.get_mode()
        if mode["mode"] == "":
            io.set_mode(mode="MD0 SEL1", direction=1, equal=0, latch=0, ch=1)
            io.set_mode(mode="MD0 SEL1", direction=1, equal=0, latch=0, ch=2)
            self.board_setting(io)
        io.initialize()
        io.reset(ch=1)
        io.set_mode("MD0", 0, 1, 0, ch=1)

        return io

    def get_dome_reading(self):
        self.dome_encoffset = self.Config.dome_encoffset
        self.dome_enc1loop = self.Config.dome_enc1loop
        self.dome_enc2arcsec = 3600.0 * 360 / self.dome_enc1loop
        self.dome_enc_tel_offset = self.Config.dome_enc_tel_offset * 360
        # counter = self.dio.get_position()
        counter = self.io.get_counter(ch=1)
        # print('self,dio.get_counter : ', counter.to_int())
        counter = int(counter)  # .to_int()
        dome_enc_arcsec = -int(
            ((counter - self.dome_encoffset) * self.dome_enc2arcsec)
            - self.dome_enc_tel_offset
        )
        while dome_enc_arcsec > 1800.0 * 360:
            dome_enc_arcsec -= 3600.0 * 360
        while dome_enc_arcsec <= -1800.0 * 360:
            dome_enc_arcsec += 3600.0 * 360
        self.dome_position = dome_enc_arcsec
        return self.dome_position

    def get_reading(self):
        cntAz = int(self.io.get_counter(unsigned=False, ch=1))
        cntEl = int(self.io.get_counter(unsigned=False, ch=2))
        """unsigned
        if cntAz < 360*3600./self.resolution:
            #encAz = (324*cntAz+295)/590
            encAz = cntAz*self.resolution
        else:
            encAz = -(2**32-cntAz)*self.resolution
            pass
        """
        encAz = cntAz * self.resolution
        Az = encAz * u.arcsec
        """ unsigned
        if cntEl < 360*3600./self.resolution:
            #encEl = (324*cntEl+295)/590
            encEl = cntEl*self.resolution
        else:
            encEl = -(2**32-cntEl)*self.resolution
            pass
        """
        encEl = cntEl * self.resolution
        El = encEl + 45 * 3600 * u.arcsec
        AzEl = {"Az": Az, "El": El}

        return AzEl  # , _utc]

    def finalize(self) -> None:
        pass

    def board_setting(self, io, z_mode=""):
        self.logger.info("Initialize start")
        io.set_z_mode(clear_condition=z_mode, latch_condition="", z_polarity=0, ch=1)
        io.set_z_mode(clear_condition=z_mode, latch_condition="", z_polarity=0, ch=2)
        print("origin setting mode : ", z_mode)
        self.logger.info("initialize end")
        return



__all__ = ["CPZ7204"]

import time

import astropy.units as u

from ... import get_logger, utils
from .motor_base import Motor


class CPZ7204(Motor):
    Manufacturer = "Interface"
    Model = "CPZ7204"

    Identifier = "rsw_id"

    def __init__(self) -> None:
        self.logger = get_logger(self.__class__.__name__)
        self.rsw_id = self.Config.rsw_id
        self.io = self._initialize_io()

    @utils.skip_on_simulator
    def _initialize_io(self):
        import pyinterface

        io = pyinterface.open(7204, self.rsw_id)
        if io is None:
            raise RuntimeError("Cannot communicate with the PCI board.")
        io.initialize()

        return io

    def set_step(self, position: str) -> None:
        self.move(position)
        return

    def get_step(self) -> str:
        status = self.io.get_status()
        if status["busy"]:
            position = "MOVE"
        elif status["limit"]["+EL"] == 0:
            position = "OUT"
        elif status["limit"]["-EL"] == 0:
            position = "IN"
        else:
            self.logger.warning("limit error")
            position = "ERROR"
        return position

    def move(self, position: str) -> None:
        if position == "OUT":
            step = 1
            self.logger.info("m4 move out")
        elif position == "IN":
            step = -1
            self.logger.info("m4 move in")
        else:
            self.logger.warning("parameter error")
            return
        self.io.set_motion(
            mode="JOG",
            acc_mode="SIN",
            low_speed=100,
            speed=1000,
            acc=500,
            step=step,
            axis=1,
        )
        self.io.start_motion(mode="JOG")
        self.logger.info("start_motion")
        time.sleep(0.5)
        return

    def stop(self, mode: str) -> None:
        # mode: "DEC", "IMMEDIATE"
        self.io.stop_motion(mode, 1)
        return

    def finalize(self) -> None:
        self.io.output_do([0, 0, 0, 0], 1)

    def set_speed(self, speed: float, axis: str) -> None:
        pass

    def get_speed(self, axis: str) -> u.Quantity:
        pass
